# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: input_file.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='input_file.proto',
  package='aihwx',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x10input_file.proto\x12\x05\x61ihwx\x1a\x0c\x63ommon.proto\"\\\n\x07\x44\x61taset\x12\x12\n\ndataset_id\x18\x01 \x02(\t\x12)\n\ntransforms\x18\x02 \x03(\x0b\x32\x15.aihwx.TransformProto\x12\x12\n\nbatch_size\x18\x03 \x01(\x05\"\x9f\x01\n\x08Training\x12(\n\toptimizer\x18\x01 \x02(\x0b\x32\x15.aihwx.OptimizerProto\x12(\n\tscheduler\x18\x02 \x01(\x0b\x32\x15.aihwx.SchedulerProto\x12/\n\rloss_function\x18\x03 \x01(\x0b\x32\x18.aihwx.LossFunctionProto\x12\x0e\n\x06\x65pochs\x18\x04 \x01(\x05\"\x95\x01\n\rTrainingInput\x12\x1f\n\x07version\x18\x01 \x02(\x0b\x32\x0e.aihwx.Version\x12\x1f\n\x07\x64\x61taset\x18\x02 \x02(\x0b\x32\x0e.aihwx.Dataset\x12\x1f\n\x07network\x18\x03 \x02(\x0b\x32\x0e.aihwx.Network\x12!\n\x08training\x18\x04 \x02(\x0b\x32\x0f.aihwx.Training'
  ,
  dependencies=[common__pb2.DESCRIPTOR,])




_DATASET = _descriptor.Descriptor(
  name='Dataset',
  full_name='aihwx.Dataset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='aihwx.Dataset.dataset_id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transforms', full_name='aihwx.Dataset.transforms', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='aihwx.Dataset.batch_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41,
  serialized_end=133,
)


_TRAINING = _descriptor.Descriptor(
  name='Training',
  full_name='aihwx.Training',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='optimizer', full_name='aihwx.Training.optimizer', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scheduler', full_name='aihwx.Training.scheduler', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='loss_function', full_name='aihwx.Training.loss_function', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='epochs', full_name='aihwx.Training.epochs', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=136,
  serialized_end=295,
)


_TRAININGINPUT = _descriptor.Descriptor(
  name='TrainingInput',
  full_name='aihwx.TrainingInput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='aihwx.TrainingInput.version', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='aihwx.TrainingInput.dataset', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='network', full_name='aihwx.TrainingInput.network', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training', full_name='aihwx.TrainingInput.training', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=298,
  serialized_end=447,
)

_DATASET.fields_by_name['transforms'].message_type = common__pb2._TRANSFORMPROTO
_TRAINING.fields_by_name['optimizer'].message_type = common__pb2._OPTIMIZERPROTO
_TRAINING.fields_by_name['scheduler'].message_type = common__pb2._SCHEDULERPROTO
_TRAINING.fields_by_name['loss_function'].message_type = common__pb2._LOSSFUNCTIONPROTO
_TRAININGINPUT.fields_by_name['version'].message_type = common__pb2._VERSION
_TRAININGINPUT.fields_by_name['dataset'].message_type = _DATASET
_TRAININGINPUT.fields_by_name['network'].message_type = common__pb2._NETWORK
_TRAININGINPUT.fields_by_name['training'].message_type = _TRAINING
DESCRIPTOR.message_types_by_name['Dataset'] = _DATASET
DESCRIPTOR.message_types_by_name['Training'] = _TRAINING
DESCRIPTOR.message_types_by_name['TrainingInput'] = _TRAININGINPUT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Dataset = _reflection.GeneratedProtocolMessageType('Dataset', (_message.Message,), {
  'DESCRIPTOR' : _DATASET,
  '__module__' : 'input_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Dataset)
  })
_sym_db.RegisterMessage(Dataset)

Training = _reflection.GeneratedProtocolMessageType('Training', (_message.Message,), {
  'DESCRIPTOR' : _TRAINING,
  '__module__' : 'input_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Training)
  })
_sym_db.RegisterMessage(Training)

TrainingInput = _reflection.GeneratedProtocolMessageType('TrainingInput', (_message.Message,), {
  'DESCRIPTOR' : _TRAININGINPUT,
  '__module__' : 'input_file_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.TrainingInput)
  })
_sym_db.RegisterMessage(TrainingInput)


# @@protoc_insertion_point(module_scope)
