# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import onnx_common_pb2 as onnx__common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='common.proto',
  package='aihwx',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0c\x63ommon.proto\x12\x05\x61ihwx\x1a\x11onnx_common.proto\"F\n\x0ePrimitiveProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"F\n\x0eTransformProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"O\n\x17\x41\x63tivationFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"m\n\nLayerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\x12)\n\nstate_dict\x18\x03 \x03(\x0b\x32\x15.aihwx.AttributeProto\"\x86\x01\n\x19LayerOrActivationFunction\x12=\n\x13\x61\x63tivation_function\x18\x01 \x01(\x0b\x32\x1e.aihwx.ActivationFunctionProtoH\x00\x12\"\n\x05layer\x18\x02 \x01(\x0b\x32\x11.aihwx.LayerProtoH\x00\x42\x06\n\x04item\"F\n\x0eOptimizerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"F\n\x0eSchedulerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"I\n\x11LossFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\";\n\x07Network\x12\x30\n\x06layers\x18\x01 \x03(\x0b\x32 .aihwx.LayerOrActivationFunction\"(\n\x07Version\x12\x0e\n\x06schema\x18\x01 \x02(\x03\x12\r\n\x05opset\x18\x02 \x02(\x03'
  ,
  dependencies=[onnx__common__pb2.DESCRIPTOR,])




_PRIMITIVEPROTO = _descriptor.Descriptor(
  name='PrimitiveProto',
  full_name='aihwx.PrimitiveProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.PrimitiveProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.PrimitiveProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42,
  serialized_end=112,
)


_TRANSFORMPROTO = _descriptor.Descriptor(
  name='TransformProto',
  full_name='aihwx.TransformProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.TransformProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.TransformProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=114,
  serialized_end=184,
)


_ACTIVATIONFUNCTIONPROTO = _descriptor.Descriptor(
  name='ActivationFunctionProto',
  full_name='aihwx.ActivationFunctionProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.ActivationFunctionProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.ActivationFunctionProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=186,
  serialized_end=265,
)


_LAYERPROTO = _descriptor.Descriptor(
  name='LayerProto',
  full_name='aihwx.LayerProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.LayerProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.LayerProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state_dict', full_name='aihwx.LayerProto.state_dict', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=267,
  serialized_end=376,
)


_LAYERORACTIVATIONFUNCTION = _descriptor.Descriptor(
  name='LayerOrActivationFunction',
  full_name='aihwx.LayerOrActivationFunction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='activation_function', full_name='aihwx.LayerOrActivationFunction.activation_function', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='layer', full_name='aihwx.LayerOrActivationFunction.layer', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='item', full_name='aihwx.LayerOrActivationFunction.item',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=379,
  serialized_end=513,
)


_OPTIMIZERPROTO = _descriptor.Descriptor(
  name='OptimizerProto',
  full_name='aihwx.OptimizerProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.OptimizerProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.OptimizerProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=515,
  serialized_end=585,
)


_SCHEDULERPROTO = _descriptor.Descriptor(
  name='SchedulerProto',
  full_name='aihwx.SchedulerProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.SchedulerProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.SchedulerProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=587,
  serialized_end=657,
)


_LOSSFUNCTIONPROTO = _descriptor.Descriptor(
  name='LossFunctionProto',
  full_name='aihwx.LossFunctionProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='aihwx.LossFunctionProto.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='arguments', full_name='aihwx.LossFunctionProto.arguments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=659,
  serialized_end=732,
)


_NETWORK = _descriptor.Descriptor(
  name='Network',
  full_name='aihwx.Network',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='layers', full_name='aihwx.Network.layers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=734,
  serialized_end=793,
)


_VERSION = _descriptor.Descriptor(
  name='Version',
  full_name='aihwx.Version',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='schema', full_name='aihwx.Version.schema', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='opset', full_name='aihwx.Version.opset', index=1,
      number=2, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=795,
  serialized_end=835,
)

_PRIMITIVEPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_TRANSFORMPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_ACTIVATIONFUNCTIONPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LAYERPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LAYERPROTO.fields_by_name['state_dict'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LAYERORACTIVATIONFUNCTION.fields_by_name['activation_function'].message_type = _ACTIVATIONFUNCTIONPROTO
_LAYERORACTIVATIONFUNCTION.fields_by_name['layer'].message_type = _LAYERPROTO
_LAYERORACTIVATIONFUNCTION.oneofs_by_name['item'].fields.append(
  _LAYERORACTIVATIONFUNCTION.fields_by_name['activation_function'])
_LAYERORACTIVATIONFUNCTION.fields_by_name['activation_function'].containing_oneof = _LAYERORACTIVATIONFUNCTION.oneofs_by_name['item']
_LAYERORACTIVATIONFUNCTION.oneofs_by_name['item'].fields.append(
  _LAYERORACTIVATIONFUNCTION.fields_by_name['layer'])
_LAYERORACTIVATIONFUNCTION.fields_by_name['layer'].containing_oneof = _LAYERORACTIVATIONFUNCTION.oneofs_by_name['item']
_OPTIMIZERPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_SCHEDULERPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_LOSSFUNCTIONPROTO.fields_by_name['arguments'].message_type = onnx__common__pb2._ATTRIBUTEPROTO
_NETWORK.fields_by_name['layers'].message_type = _LAYERORACTIVATIONFUNCTION
DESCRIPTOR.message_types_by_name['PrimitiveProto'] = _PRIMITIVEPROTO
DESCRIPTOR.message_types_by_name['TransformProto'] = _TRANSFORMPROTO
DESCRIPTOR.message_types_by_name['ActivationFunctionProto'] = _ACTIVATIONFUNCTIONPROTO
DESCRIPTOR.message_types_by_name['LayerProto'] = _LAYERPROTO
DESCRIPTOR.message_types_by_name['LayerOrActivationFunction'] = _LAYERORACTIVATIONFUNCTION
DESCRIPTOR.message_types_by_name['OptimizerProto'] = _OPTIMIZERPROTO
DESCRIPTOR.message_types_by_name['SchedulerProto'] = _SCHEDULERPROTO
DESCRIPTOR.message_types_by_name['LossFunctionProto'] = _LOSSFUNCTIONPROTO
DESCRIPTOR.message_types_by_name['Network'] = _NETWORK
DESCRIPTOR.message_types_by_name['Version'] = _VERSION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PrimitiveProto = _reflection.GeneratedProtocolMessageType('PrimitiveProto', (_message.Message,), {
  'DESCRIPTOR' : _PRIMITIVEPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.PrimitiveProto)
  })
_sym_db.RegisterMessage(PrimitiveProto)

TransformProto = _reflection.GeneratedProtocolMessageType('TransformProto', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFORMPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.TransformProto)
  })
_sym_db.RegisterMessage(TransformProto)

ActivationFunctionProto = _reflection.GeneratedProtocolMessageType('ActivationFunctionProto', (_message.Message,), {
  'DESCRIPTOR' : _ACTIVATIONFUNCTIONPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.ActivationFunctionProto)
  })
_sym_db.RegisterMessage(ActivationFunctionProto)

LayerProto = _reflection.GeneratedProtocolMessageType('LayerProto', (_message.Message,), {
  'DESCRIPTOR' : _LAYERPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.LayerProto)
  })
_sym_db.RegisterMessage(LayerProto)

LayerOrActivationFunction = _reflection.GeneratedProtocolMessageType('LayerOrActivationFunction', (_message.Message,), {
  'DESCRIPTOR' : _LAYERORACTIVATIONFUNCTION,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.LayerOrActivationFunction)
  })
_sym_db.RegisterMessage(LayerOrActivationFunction)

OptimizerProto = _reflection.GeneratedProtocolMessageType('OptimizerProto', (_message.Message,), {
  'DESCRIPTOR' : _OPTIMIZERPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.OptimizerProto)
  })
_sym_db.RegisterMessage(OptimizerProto)

SchedulerProto = _reflection.GeneratedProtocolMessageType('SchedulerProto', (_message.Message,), {
  'DESCRIPTOR' : _SCHEDULERPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.SchedulerProto)
  })
_sym_db.RegisterMessage(SchedulerProto)

LossFunctionProto = _reflection.GeneratedProtocolMessageType('LossFunctionProto', (_message.Message,), {
  'DESCRIPTOR' : _LOSSFUNCTIONPROTO,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.LossFunctionProto)
  })
_sym_db.RegisterMessage(LossFunctionProto)

Network = _reflection.GeneratedProtocolMessageType('Network', (_message.Message,), {
  'DESCRIPTOR' : _NETWORK,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Network)
  })
_sym_db.RegisterMessage(Network)

Version = _reflection.GeneratedProtocolMessageType('Version', (_message.Message,), {
  'DESCRIPTOR' : _VERSION,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:aihwx.Version)
  })
_sym_db.RegisterMessage(Version)


# @@protoc_insertion_point(module_scope)
